#!/usr/bin/env python -- # -*- python -*-
# $Id: ratio.py,v 1.6 2000-04-23 13:16:47-05 annis Exp annis $
# $Source: /export/home/annis/MA/RCS/ratio.py,v $
#
# Copyright (c) 1999, 2000 William S. Annis.  All rights reserved.
# This is free software; you can redistribute it and/or modify it
# under the same terms as Perl (the Artistic Licence).

"""Ratio mathematics.

These aren't quite your normal ratios.  This class is only useful for
people interested in dealing with musical ratios... just intonation
fanatics.  In particular, the definition of addition and subtraction
here would be surprising to your elementary school math teacher.
"""

from types import ListType, TupleType
import math, string

# I do a lot of tedious, repetitive math in this module.  Let's eat
# memory, not time.
class memoize:
    def __init__(self, fn):
        self.fn = fn
        self.args = {}

    def __call__(self, *args):
        if not self.args.has_key(args):
            self.args[args] = apply(self.fn, args)

        return self.args[args]


# For building just chords (often outside the scale)
CHORDS = {'major': (4, 5, 6),
          'minor': (10, 12, 15),
          'diminished': (5, 6, 7),
          'subminor': (6, 7, 9),
          'dom-7': (4, 5, 6, 7),
          '1-3-7': (4, 6, 7),           # Nameless consonance.
          'major-7': (8, 10, 12, 15),
          'minor-7': (10, 12, 15, 18),
          'subminor-7': (6, 7, 9, 11),  # Dreaded 11-limit tessaract infection!
          'dom-9': (4, 5, 6, 7, 9),
          'major-9': (8, 10, 12, 15, 18),
          'minor-9': (20, 24, 30, 36, 45)
          }

# Now, generate a map of identities for the chords above.  Shove into
# a database keyed on the sorted identities.  This is useful for generating
# lists of available harmonic material in a system.
def ident(n):
    while n % 2 == 0:
        n = n / 2

    return n

def _concords():
    cs = {}
    for name, fs in CHORDS.items():
        g = map(ident, fs)
        g.sort()
        cs[tuple(g)] = name

    return cs
CONCORDS = _concords()

def _gcd_(a, b):
    """Compute the greatest common divisor by Euclid's Algorithm."""
    while b != 0:
        a, b = b, a % b

    return a
# Now, memoize finding the GCD.
_gcd = memoize(_gcd_)


class Ratio:
    def __init__(self, num, denom):
        self.num, self.denom = num, denom

    def __repr__(self):
        return "<%s:%s>" % (self.num, self.denom)

    def __add__(self, r):
        return Ratio(self.num * r.num, self.denom * r.denom).reduce()

    def __sub__(self, r):
        return Ratio(self.num * r.denom, self.denom * r.num).reduce()

    def __float__(self):
        return self.num / (self.denom * 1.0)

    def __cmp__(self, r):
        if self.num == r.num and self.denom == r.denom:
            return 0
        else:
            return cmp(float(self), float(r))

    def __hash__(self):
        return hash(`self`)

    def complement(self, against=None):
        if against:
            return against - self
        else:
            return Ratio(2, 1) - self

    def cents(self):
        """R.cents() - return the ratio value in cents"""
        return math.log10(float(self)) * 3986.31371386

    def reduce(self):
        """R.reduce() - simplify this ratio"""
        g = _gcd(self.num, self.denom)
        self.num = self.num / g
        self.denom = self.denom / g
        return self

    def octave(self):
        """R.octave() - shift ratio into the range 1:1 =< x <= 2:1"""
        while float(self) > 2.0:
            new = self - Ratio(2, 1)
            self.num, self.denom = new.num, new.denom

        while float(self) < 1.0:
            new = self + Ratio(2, 1)
            self.num, self.denom = new.num, new.denom

        return self

    def katapykne(self, n=1):
        """Do n divisions of the ratio interval by katapyknosis.

        If n==1, the (non-superparticular) ratio is simply divided.
        So, <5:3> will return [<5:4>, <4:3>].

        This does 1:1 katapyknosis (obviously).
        """
        ints = range(self.num * n, self.denom * n - 1, -1)

        intervals = []
        for i in range(len(ints) - 1):
            intervals.append(Ratio(ints[i], ints[i + 1]))

        return intervals

    # Allow 1/2 or 2/1 kp for those who find 1/1 boring.  Formula
    # from Chalmers' "Divisions of the Tetrachord," p. 13.
    def katapykne_ab(self, a, b):
        return [Ratio((a + b) * self.num, (b * self.num + a * self.denom)),
                Ratio((b * self.num + a * self.denom), (a + b) * self.denom)]

    # Now you can say r=Ratio(4, 3); r[1:1], r[2:1], etc.
    def __getslice__(self, a, b):
        return self.katapykne_ab(a, b)

    def chord_of(self, chord_name, ident=0):
        """r.chord_of(chord, {ident}) - generate a chord

        Return a list of ratios which make a chord of the named type.
        The types available: major, minor, diminished, subminor, dom-7,
        major-7, minor-7, subminor-7, dom-9, major-9, minor-9.  See
        the source for a complete list.

        The default is to make the instance ration the lowest identity
        of the chord.  If you wanted to make a septimal minor (or
        subminor) triad with 3:2 as the second note, you'd do this:

            >>> Ratio(3, 2).chord_of('subminor', 1)
            [<9:7>, <3:2>, <27:14>]

        Since Python is zero-indexed, '1' is the second identity in the
        chord.

        Finally, you may use a list or tuple of harmonic identities
        instead of a chord name:

            >>> Ratio(1, 1).chord_of((5, 7, 9, 33), 1)
            [<10:7>, <1:1>, <9:7>, <33:28>]
        """
        if type(chord_name) in [ListType, TupleType]:
            harmonics = chord_name
        else:
            if not CHORDS.has_key(chord_name):
                raise Exception, "no such chord type: %s" % chord_name
            harmonics = CHORDS[chord_name]

        chord = []
        base = self - Ratio(harmonics[ident], 1)
        for h in harmonics:
            int = base + Ratio(h, 1)
            int.octave()
            int.reduce()
            chord.append(int)

        return chord


class Duodene:
    """Duodene Chart

    This is a simple chart creation class which produces things like
    Ellis's duodenes.  The default Duodene produces one for the 4:5:6
    triad.

    This does a fine job of producing 2D lattice diagrams.
    """
    def __init__(self, hrat=None, hn=4, vrat=None, vn=3):
        """Generate a new duodene.

        d = Duodene(horizontal_ratio, horizontal_number, vert_ratio, vert_num)
        d = Duodene(vrat=Ratio(7, 4), hn=5, vn=5)

        The default is for the horizontal scale to be 4 3:2 (perfect 5ths),
        and for the vertical scale to be 3 5:4 (major 3rds).  The chart
        will be centered on 1:1, but is biased toward producing more ratios
        above 1:1, rather than below.
        """
        self.hn = hn
        self.vn = vn

        if not hrat:
            hrat = Ratio(3, 2)
        if not vrat:
            vrat = Ratio(5, 4)

        base = Ratio(1, 1)
        self.duodene = {}
        h = - hn / 2 + 1
        for hi in range(hn):
            hbase = tune(hrat, h)
            #print "hbase", hbase
            v = - vn / 2 + 1
            for vj in range(vn):
                #print hbase + tune(vrat, v), vrat, v, tune(vrat, v)
                self.duodene[(hi, vj)] = hbase + tune(vrat, v)
                self.duodene[(hi, vj)].octave()
                
                v = v + 1
            h = h + 1

    def __repr__(self):
        repr = ""
        for vi in range(self.vn - 1, -1, -1):
            for hj in range(self.hn):
                repr = repr + "%10s" % self.duodene[(hj, vi)]
            repr = repr + "\n"

        return repr

    def gamut(self):
        gam = []
        for loc, ratio in self.duodene.items():
            gam.append(ratio)
        
        gam.sort()
        return gam
        

# A few utilities...
def cents(intervals):
    """Convert a sequence of scale degrees in a sequence of cent values."""
    centt = []
    for i in intervals:
        centt.append(i.cents())

    return centt

def _tune(ratio, n):
    """Tune strings by some ratio.

    This returns the +/- Nth ratio resulting from tuning by an interval.
    That is, tune by fifths, say.
    """
    base = Ratio(1, 1)
    if n < 0:
        n = n * -1
        for i in range(n):
            base = base - ratio

        return base.octave()
    else:
        for i in range(n):
            base = base + ratio

        return base.octave()
tune = memoize(_tune)

def intervals2gamut(intervals):
    """Convert a sequence of intervals into a sequence of scale degrees."""
    acc = Ratio(1, 1)
    ratios = []
    for interval in intervals:
        ratios.append(acc)
        acc = acc + interval
    ratios.append(acc)

    return ratios

def gamut2intervals(ratios):
    """Convert a sequence of scale degrees into a sequence of intervals."""
    ints = []
    for i in range(1, len(ratios)):
        ints.append(ratios[i] - ratios[i - 1])

    return ints

def gamut2series(gamut):
    """Turn a sequence of pitches into harmonic overtone list.

    This gets a touch confused when the series spans more than an octave.
    You may need to fiddle with the argument to guarantee a completely
    reduced answer.
    """
    # Blunt, blunt, blunt.
    cd = 1
    for g in gamut:
        cd = cd * g.denom

    overtones = []
    for g in gamut:
        overtones.append(g.num * (cd / g.denom))

    # Find the GCD of all these numbers.
    gcd = _gcd(overtones[0], overtones[1])
    for r in overtones[2:]:
        gcd = _gcd(gcd, r)

    # Now divide by the gcd, but don't bother if it's 1.
    if gcd != 1:
        for i in range(len(overtones)):
            overtones[i] = overtones[i] / gcd

    return overtones

def gamut2identities(gamut):
    """Turn a sequence of pitches into harmonic overtone identity list."""
    series = gamut2series(gamut)

    idents = []
    for s in series:
        ident = s
        while ident % 2 == 0:
            ident = ident / 2

        idents.append(ident)

    return idents

# This could be a lot prettier.  I'll have to look into ways to most
# efficiently use space, since more complex scales will roll over the edge.
# I should get something much like this (alligned on the :)
#>>> ms = [Ratio(1, 1), Ratio(28, 27), Ratio(9, 8), Ratio(4, 3), Ratio(3, 2), Ratio(14, 9), Ratio(27, 16)]
#>>> ratiotable(ms)
#<1:1>   |    <1:1>   <28:27>   <9:8>    <4:3>    <3:2>   <14:9>   <27:16>
#<28:27> |   <27:28>   <1:1>  <243:224>  <9:7>   <81:56>   <3:2>  <729:448>
#<9:8>   |    <8:9>  <224:243>  <1:1>   <32:27>   <4:3>  <112:81>   <3:2>
#<4:3>   |    <3:4>    <7:9>   <27:32>   <1:1>    <9:8>    <7:6>   <81:64>
#<3:2>   |    <2:3>   <56:81>   <3:4>    <8:9>    <1:1>   <28:27>   <9:8>
#<14:9>  |    <9:14>   <2:3>   <81:112>  <6:7>   <27:28>   <1:1>  <243:224>
#<27:16> |   <16:27> <448:729>  <2:3>   <64:81>   <8:9>  <224:243>  <1:1>
def ratiotable(ratios):
    """Generate a table of inter-ratio ratios.

    Assumes a scale of ratios in relation to 1/1, not inter-note.
    """
    r = ratios[:]
    r.sort()

    l = len(r)
    table = [1] * l
    for i in range(l):
        table[i] = [1] * l
        for j in range(l):
            table[i][j] = r[i] - r[j]
            table[i][j].reduce()
            # Chop off the brackets.
            table[i][j] = `table[i][j]`[1:-1]

    # Now, we try to most efficiently use our space.
    mspace = [1] * l
    for i in range(l):
        ms = 0
        for j in range(l):
            ms = max(ms, len(table[j][i]))

        # This creates a format string in the form, eg., " %5s".
        mspace[i] = " %%%is" % ms

    for i in range(l):
        print mspace[0] % `r[i]`[1:-1], "|",
        for j in range(l):
            print mspace[j] % table[j][i],
        print

# These are both a little blunt.  Only the memoizing of the GCD function
# keeps them from running much longer.
def triads(scale):
    """Generate a list of known triads in a scale."""
    n = len(scale)
    chords = {}
    concords = CONCORDS.keys()
    for i in range(n):
        for j in range(i, n):
            for k in range(j, n):
                chord = gamut2identities([scale[i], scale[j], scale[k]])
                chord.sort()
                chord = tuple(chord)
                if chord in concords:
                    chords[(scale[i], scale[j], scale[k])] = CONCORDS[chord]

    return chords

def sevenths(scale):
    """Generate a list of known four-element chords in a scale."""
    n = len(scale)
    chords = {}
    concords = CONCORDS.keys()
    for i in range(n):
        for j in range(i, n):
            for k in range(j, n):
                for l in range(k, n):
                    chord = gamut2identities([scale[i], scale[j], scale[k], scale[l]])
                    chord.sort()
                    chord = tuple(chord)
                    if chord in concords:
                        chords[(scale[i], scale[j], scale[k], scale[l])] = CONCORDS[chord]

    return chords


# EOF
